"""
IDA - image decoding algorithms
https://github.com/gejirz/image-decoding-algorithms - you can visit my repository on github
Made by Victor Andrievskiy, e-mail: superavb222@gmail.com
"""

from sys import path
from pathlib import Path
path.append(str(Path(__file__).parent.resolve()))

from IDA_main import read_image

__version__ = "0.0.4"
